//
//  SwiftAidlabSdkPlugin.swift
//  AidlabSDK
//
//  Created by Szymon Gęsicki on 05/02/2021.
//  Copyright © 2020-2024 Aidlab. All rights reserved.
//

import Flutter
import UIKit

class AidlabManagerHandler: AidlabManager_, AidlabManagerDelegate {

    private let messenger: FlutterBinaryMessenger
    private lazy var aidlabManager: AidlabManager = {
        return AidlabManager(delegate: self)
    }()
    private let aidlabManagerDelegate_: AidlabManagerDelegate_
    private var discoveredDevices: [UUID: Device] = [:]
    private var connectedDevice: DeviceHandler?

    init(messenger: FlutterBinaryMessenger) {

        self.messenger = messenger
        aidlabManagerDelegate_ = AidlabManagerDelegate_(binaryMessenger: messenger)
        AidlabManager_Setup.setUp(binaryMessenger: messenger, api: self)
    }

    deinit {

        AidlabManager_Setup.setUp(binaryMessenger: messenger, api: nil)
    }

    public func didDiscover(_ device: Device) {
        discoveredDevices[device.address] = device
        aidlabManagerDelegate_.didDiscover(device: device.toFlutter(), rssi: device.rssi.int64Value, completion: { _ in })
    }

    func scan(mode: ScanMode_) throws {

        aidlabManager.scan()
    }

    func stopScan() throws {

        aidlabManager.stopScan()
    }

    func connect(device: Device_) throws {

        guard let uuid = UUID(uuidString: device.address) else {
            print("Incorrect UUID")
            return
        }

        let device = self.discoveredDevices[uuid]

        guard let device else {
            print("Device not found")
            return
        }

        connectedDevice = DeviceHandler(messenger: messenger, device: device)
        connectedDevice?.connect()
    }

    func collect(device: Device_, dataTypes: [FlutterDataType_], dataTypesToStore: [FlutterDataType_]) throws {

        guard let uuid = UUID(uuidString: device.address) else {
            print("Incorrect UUID")
            return
        }

        let device = self.discoveredDevices[uuid]

        guard let device else {
            print("Device not found")
            return
        }

        device.collect(dataTypes: dataTypes.map { DataType(rawValue: $0.signal!.rawValue)! }, dataTypesToStore: dataTypesToStore.map { DataType(rawValue: $0.signal!.rawValue)! })
    }

    func disconnect(device: Device_) throws {
        connectedDevice?.disconnect()
    }
}

public class SwiftAidlabSdkPlugin: NSObject, FlutterPlugin {

    private static var aidlabManagerHandler: AidlabManagerHandler?

    public static func register(with registrar: FlutterPluginRegistrar) {

        let channel = FlutterMethodChannel(name: "aidlab_sdk", binaryMessenger: registrar.messenger())
        registrar.addMethodCallDelegate(SwiftAidlabSdkPlugin(), channel: channel)

        aidlabManagerHandler = AidlabManagerHandler(messenger: registrar.messenger())
    }
    
    public func detachFromEngine(for registrar: FlutterPluginRegistrar) {
        
        SwiftAidlabSdkPlugin.aidlabManagerHandler = nil
    }
}
